#ifndef seleneLiftAxis_H
#define seleneLiftAxis_H
#include "seleneAngleAxis.h"
#include "seleneOffsetAxis.h"
#include "turboPmacAxis.h"
#include <array>

// Forward declaration of the seleneGuideController class to resolve the cyclic
// dependency. See https://en.cppreference.com/w/cpp/language/class.
class seleneGuideController;

/**
 * @brief Virtual axis for setting the height of the center of a Selene guide
 *
 * Please see README.md for a detailed explanation.
 */
class seleneLiftAxis : public turboPmacAxis {
  public:
    static const int numAxes_ = 6;

    /**
     * @brief Construct a new seleneLiftAxis object
     *
     * The offset axes need to be constructed in advance and they need to be
     * available by the given indices, otherwise the constructor will terminate
     * the IOC (configuration error).
     *
     * @param pC            Pointer to the associated controller
     * @param axis1No       Index of the first associated offset axis
     * @param axis2No       Index of the second associated offset axis
     * @param axis3No       Index of the third associated offset axis
     * @param axis4No       Index of the fourth associated offset axis
     * @param axis5No       Index of the fifth associated offset axis
     * @param axis6No       Index of the sixth associated offset axis
     * @param liftAxisNo    Index of the lift axis itself
     * @param angleAxisNo   Index of the angle axis itself
     */
    seleneLiftAxis(seleneGuideController *pC, int axis1No, int axis2No,
                   int axis3No, int axis4No, int axis5No, int axis6No,
                   int liftAxisNo, int angleAxisNo);

    /**
     * @brief Implementation of the `stop` function from asynMotorAxis
     *
     * Stops all underlying physical axes at once
     *
     * @param acceleration        Acceleration ACCEL from the motor record. This
     * value is currently not used.
     * @return asynStatus
     */
    asynStatus stop(double acceleration);

    /**
     * @brief Implementation of the `doPoll` function from sinqAxis.
     *
     * Some paramLib entries of the associated angle and offset axes are set in
     * this function.
     *
     * @param moving
     * @return asynStatus
     */
    asynStatus doPoll(bool *moving);

    /**
     * @brief Lift the rotation center of the guide to the specified position.
     *
     * @param position
     * @param relative
     * @param min_velocity
     * @param maxOffset_velocity
     * @param acceleration
     * @return asynStatus
     */
    asynStatus doMove(double position, int relative, double min_velocity,
                      double maxOffset_velocity, double acceleration);

    /**
     * @brief Reset all errors of the underlying physical axes.
     *
     * @param on
     * @return asynStatus
     */
    asynStatus doReset();

    /**
     * @brief This function does nothing, since this axis cannot be enabled /
     * disabled
     *
     * @param on
     * @return asynStatus
     */
    asynStatus enable(bool on);

    /**
     * @brief "Home" the virtual axes by driving all underlying real axes to
     * position 0.
     *
     * @param minVelocity
     * @param maxVelocity
     * @param acceleration
     * @param forwards
     * @return asynStatus
     */
    asynStatus doHome(double minVelocity, double maxVelocity,
                      double acceleration, int forwards);

    /**
     * @brief Write the encoder type in the corresponding PV
     *
     * Since this axis is a virtual one, it does not have an encoder. In order
     * to enable reference drives (which in case of this axis simply sets all
     * physical axes to their respective physical position zero), this function
     * sets the encoder type to "Incremental Encoder".
     *
     * @return asynStatus
     */
    asynStatus readEncoderType();

    /**
     * @brief This function does nothing, since this axis does not have an
     * encoder.
     *
     * @return asynStatus
     */
    asynStatus rereadEncoder() { return asynSuccess; }

    /**
     * @brief Start a combined movement from a virtual axis.
     *
     * This functions sets `virtualAxisMovement_` to `true` and then calls the
     * `startCombinedMove()` method.
     *
     * @return asynStatus
     */
    asynStatus startCombinedMoveFromVirtualAxis() {
        virtualAxisMovement_ = true;
        return startCombinedMove();
    }

    /**
     * @brief Start a combined movement from a real offset axis.
     *
     * This functions sets `virtualAxisMovement_` to `false` and then calls the
     * `startCombinedMove()` method.
     *
     * @return asynStatus
     */
    asynStatus startCombinedMoveFromOffsetAxis() {
        virtualAxisMovement_ = false;
        return startCombinedMove();
    }

    /**
     * @brief Start a combined movement using the target positions of the offset
     * and virtual axes.
     *
     * This function takes the target positions of the `seleneLiftAxis`, the
     * `seleneAngleAxis` and the associated `seleneOffsetAxis`' and calculates
     * the absolute positions of the underlying real axes. In pseudo-code, this
     * calculation looks somewhat like this:
     *
     * ```
     * for i in 0:1:numAxes_
     *    pos(i) = target_lift + tan(target_angle) * (x(i) - x_lift) + target(i)
     *```
     * where `x(i)` is the distance between the first axis and the i-th axis.
     * All numAxes_ positions are calculated and send to the controller as one
     *command (even though the position of one or multiple axes might not change
     *at all).
     *
     * @return asynStatus
     */
    asynStatus startCombinedMove();

    /**
     * @brief Readout of some values from the controller at IOC startup
     *
     * The following steps are performed:
     * - Read out the motor status, motor position, velocity and acceleration
     * from the MCU and store this information in the parameter library.
     * - Set the enable PV according to the initial status of the axis.
     *
     * @return asynStatus
     */
    asynStatus init();

    /**
     * @brief Calculate lift and angle axis position based on the actual
     positions of the underlying motors.

    This function implements the lift and angle calculation described in
    README.md based on the given position values of the central physical axes.
    Usually, these positions are those of axis 3 and 4.
     *
     * @param pos1
     * @param pos1
     * @param lift
     * @param angle
     */
    void deriveLiftAndAngle(double pos1, double pos2, double *lift,
                            double *angle);

    /**
     * @brief Set the offsets of axis 3 and 4 to zero and recalculate all other
     * axes positions based on it.
     *
     * Calling this function "normalizes" the lift axis, the angle axis and the
     * offset axes by performing a "normalization" operation. For details, see
     * README.md. This function does not communicate with the controller.
     *
     * @return asynStatus
     */
    asynStatus normalizePositions();

    /**
     * @brief Calculate and return the positions of all 6 underlying physical
     * axes from the given lift and angle without considering the offset axes.
     *
     * In order to calculate the positions, this function convertes the given
     * lift and angle into a linear equation. The x-values of the axes are read
     * out from the controller in the constructor, which allows us to calculate
     * the corresponding y-values. These are then corrected by the zero offset
     * of the corresponding axis in order to calculate the actual positon.
     *
     * The corresponding equation is:
     * ```
     * y = lift + tan(angle)*(x-xLift_)
     * ```
     *
     * This function does not change any values in the parameter library and
     * does not communicate with the controller.
     *
     * @param lift    Position of the virtual lift axis in mm
     * @param angle   Angle of the virtual angle axis in radian
     * @return std::array<double, 6>
     */
    std::array<double, numAxes_> positionsFromLiftAndAngle(double lift,
                                                           double angle);

    /**
     * @brief Access the angle axis associated with this lift axis.
     *
     * @return seleneAngleAxis*
     */
    seleneAngleAxis *angleAxis() { return angleAxis_; }

    /**
     * @brief Checks whether the current movement was triggered from the lift /
     * angle axes or the offset axes.
     *
     * @return true   Triggered by lift / angle axis
     * @return false  Triggered by an offset axis
     */
    bool virtualAxisMovement() { return virtualAxisMovement_; }

    /**
     * @brief Read the target position from the axis and write it into the
     * provided pointer.
     *
     * @param targetPosition
     * @return asynStatus
     */
    asynStatus targetPosition(double *targetPosition);

    /**
     * @brief Access the position of the lift axis
     *
     */
    double xPos() { return xLift_; }

    /**
     * @brief Return a pointer to the associated offset axis with index "idx"
     *
     * If the given index is out of bounds, return a nullptr instead.
     *
     * @param index
     * @return seleneOffsetAxis*
     */
    seleneOffsetAxis *offsetAxis(int idx);

  protected:
    std::array<seleneOffsetAxis *, numAxes_> offsetAxes_;

    seleneGuideController *pC_;
    seleneAngleAxis *angleAxis_;

    // Horizontal position of the virtual lift axis
    double xLift_;

    // True, if the target has been set from this axis
    bool targetSet_;

    // True, if the movement has been started from the lift or angle axis
    bool virtualAxisMovement_;
};

#endif
